import importlib
import importlib.util
import os
from datetime import datetime
from urllib.parse import urlparse, urlencode, parse_qsl

from random_user_agent.user_agent import UserAgent
from random_user_agent.params import SoftwareName, OperatingSystem


PROJECT_FOLDER = os.path.join(os.getcwd(), 'projects')

_software_names = [SoftwareName.CHROME.value]
_operating_systems = [OperatingSystem.WINDOWS.value, OperatingSystem.LINUX.value]
UAFactory = UserAgent(software_names=_software_names, operating_systems=_operating_systems, limit=100)


def get_headers(language: str):
    """ Constructs request headers with given language header and random user-agent """
    return {
        'User-Agent': UAFactory.get_random_user_agent(),
        'Accept-Language': language,
        'Accept-Encoding': 'deflate, gzip;q=1.0, *;q=0.5',
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8'
    }


# Mostly copied from scrapy.contrib.linkextractor
IGNORED_EXTENSIONS = [
    # images
    'mng', 'pct', 'bmp', 'gif', 'jpg', 'jpeg', 'png', 'pst', 'psp', 'tif',
    'tiff', 'ai', 'drw', 'dxf', 'eps', 'ps', 'svg',

    # audio
    'mp3', 'wma', 'ogg', 'wav', 'ra', 'aac', 'mid', 'au', 'aiff',

    # video
    '3gp', 'asf', 'asx', 'avi', 'mov', 'mp4', 'mpg', 'qt', 'rm', 'swf', 'wmv', 'm4a',

    # other
    'css', 'pdf', 'doc', 'docx', 'xls', 'xlsx', 'dmg', 'exe', 'bin', 'rss', 'zip', 'rar',
]


def norm_url(url: str):
    parsed = urlparse(url)
    # Sort query parameters if there are any
    query = '?' + urlencode(sorted(parse_qsl(parsed.query))) if parsed.query else ''
    # Drop fragments and rebuild URL
    return f'{parsed.scheme}://{parsed.netloc}{parsed.path if parsed.path.startswith("/") else f"/{parsed.path}"}{query}'


def get_first_or_list(from_result):
    """ Return the first element, if there's only one, otherwise returns the whole list """
    return from_result[0] if (type(from_result) == list and len(from_result) == 1) else from_result


def remove_multi_whitespace(string_or_list):
    """ Cleans redundant whitespace from extracted data """
    if type(string_or_list) == str:
        return ' '.join(string_or_list.split())
    return [' '.join(string.split()) for string in string_or_list]


def load_project(project_name, project_folder=None):
    dir_path = project_folder or PROJECT_FOLDER
    path = os.path.join(dir_path, project_name + '.py')
    spec = importlib.util.spec_from_file_location(project_name, path)
    project = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(project)
    return project




import asyncio
import json
import logging
import time
from typing import Callable, List, Dict, Any, Union

from urllib.parse import urlparse

import prettytable
import completely
from aiohttp import AsyncResolver, TCPConnector, ClientSession
from diskcache import Index
from lxml import html as DOMParser
from parsel import Selector

from microwler.page import Page
from microwler.settings import Settings
from microwler import utils

LOG = logging.getLogger(__name__)


class Microwler:
    """
    Each `Microwler` targets exactly one domain/website.
    """

    def __init__(self,
                 start_url: str,
                 select: Dict[str, Union[str, Callable[[Selector], Any]]] = None,
                 transform: Callable[[dict], dict] = None,
                 settings: dict = None):
        """
        Setup a new `Microwler` instance
        Arguments:
            start_url: the URL to start crawling
            select: A `dict` with *selectors* [(read more)](/microwler/scraping/#selectors)
            transform: Function to transform scraped data after crawling ([read more](/microwler/faq/#what-are-transformers))
            settings: A `dict` with configuration parameters for the crawler [(read more)](/microwler/configuration/#settings)
        """
        self.start_url = start_url
        parsed = urlparse(start_url)
        self._domain = parsed.netloc
        self._base_url = f'{parsed.scheme}://{self._domain}{parsed.path}'
        self._selectors = select
        self._transformer = transform
        self._settings = Settings(settings)
        self._seen_urls = set()
        self._session: Union[ClientSession, None] = None
        self._limiter = asyncio.BoundedSemaphore(self._settings.max_concurrency)
        self._verbose = False
        self._errors = dict()
        self._results = dict()
        self.set_cache()

    def set_cache(self, force=False):
        if self._settings.caching or force:
            self._cache = Index(f'./.microwler/cache/{self._domain}')
        else:
            self._cache = None

    async def _get(self, url):
        async with self._limiter:
            try:
                heads = utils.get_headers(self._settings.language)
                async with self._session.get(url, timeout=15, headers=heads) as response:
                    text = await response.text()
                    if self._verbose:
                        LOG.info(f'Processed: {url} [{response.status}]')
                    return text, response.status
            except TimeoutError:
                if self._verbose:
                    LOG.warning(f'Timeout error: {url}')
                return None, None

    def _find_links(self, html):
        """ Extract relevant links from an HTML document """
        dom = DOMParser.fromstring(html)
        dom.make_links_absolute(self._base_url)
        links = {
            link for link in dom.xpath(self._settings.link_filter)
            if link.startswith(self._base_url)  # stay on this website
            if link not in self._results  # try to filter global duplicates in order to avoid extra loop steps later
            and not any([link.lower().endswith(e) for e in utils.IGNORED_EXTENSIONS])  # ignore file extensions
        }
        return list(links)

    async def _get_one(self, url):
        try:
            text, status = await self._get(url)
            if text is None and status is None:
                self._errors[url] = 'Timeout Error'
            else:
                links = self._find_links(text)
                return url, status, text, links
        except Exception as e:
            if self._verbose:
                LOG.error(f'Download error: {e} [{url}]')
            self._errors[url] = str(e)
            return None

    async def _get_batch(self, to_fetch):
        futures, results = [], []
        for url in to_fetch:
            normalized_url = utils.norm_url(url)
            if normalized_url in self._results:
                continue
            if self._settings.delta_crawl:
                if normalized_url in self._cache:
                    if self._verbose:
                        LOG.info(f'Dropped pre-cached URL [{normalized_url}]')
                    continue
            self._results[normalized_url] = None
            futures.append(self._get_one(normalized_url))

        for future in asyncio.as_completed(futures):
            try:
                result = (await future)
                if result is not None:
                    results.append(result)
            except Exception as e:
                LOG.error(f'Error while crawling: {e}')
                exit(1)
        return results

    async def _crawl(self, loop):
        LOG.info(f'Crawler started [{self._domain}]')
        resolver = AsyncResolver(nameservers=self._settings.dns_providers)
        tcpc = TCPConnector(resolver=resolver)
        self._session = ClientSession(loop=loop, connector=tcpc)
        pipeline = [self.start_url]
        try:
            for depth in range(self._settings.max_depth + 1):
                batch = await self._get_batch(pipeline)
                pipeline = []
                for url, status, text, links in batch:
                    pipeline.extend(links)
                    page = Page(url, status, depth, links, text)
                    self._results[url] = page
        finally:
            await self._session.close()
            LOG.info(f'Crawler stopped [{self._domain}]')

    def _process(self, sort_urls=False, keep_source=False):
        if sort_urls:
            LOG.info(f'Sorting results ... [{self._domain}]')
            self._results = {url: self._results[url] for url in sorted(self._results)}

        if self._selectors:
            LOG.info(f'Extracting data ... [{self._domain}]')
            for url, page in self._results.items():
                self._results[url] = page.scrape(self._selectors, keep_source=keep_source)

                if self._transformer is not None:
                    self._results[url] = page.transform(self._transformer)
        count = len(self._settings.exporters)
        if count:
            LOG.info(f'Exporting to {count} destinations... [{self._domain}]')
            for exporter_cls in self._settings.exporters:
                instance = exporter_cls(self._domain, list(self._results.values()), self._settings)
                instance.export()

        if self._cache is not None:
            LOG.info(f'Caching results ... [{self._domain}]')
            for page in self._results.values():
                if page.url not in self._errors:
                    self._cache[page.url] = page.__dict__

    def run(self, verbose: bool = False, sort_urls: bool = False, keep_source: bool = False):
        """
        Starts the crawler instance. You can retrieve the results using the `crawler.data` or `crawler.pages` properties.
        Arguments:
            verbose: log progress to `stdout` while crawling
            sort_urls: sort result list by URL
            keep_source: per default, the page content will be after scraping
        """

        self._verbose = verbose
        start = time.time()
        LOG.info('Starting engine ...')
        loop = asyncio.get_event_loop()
        future = asyncio.Task(self._crawl(loop=loop))
        loop.run_until_complete(future)
        loop.close()
        crawl_time = time.time() - start

        if len(self._results):
            self._process(sort_urls=sort_urls, keep_source=keep_source)
            total_time = time.time() - start
            table = prettytable.PrettyTable()
            table.add_column('Pages', [len(self._results)])
            table.add_column('Crawl Time', [f'{round(crawl_time, 2)}s'])
            table.add_column('Crawl Speed', [f'{round(len(self._results) / crawl_time, 2)} p/s'])
            table.add_column('Errors', [len(self._errors)])
            if len(self._results):
                table.add_column('Processing Time', [f'{round(total_time - crawl_time, 2)}s'])
                if self._selectors:
                    table.add_column('Data Completeness', [f'{int(completely.measure(self.results) * 100)}%'])
            table.add_column('Total Time', [f'{round(total_time, 2)}s'])
            print(table)

    async def run_async(self, event_loop, sort_urls: bool = False, keep_source: bool = False):
        """
        For running the crawler from another `asyncio` app, which has an existing event loop, i.e. Quart.
        Arguments:
             event_loop: existing event loop, i.e. as a result of asyncio.get_event_loop()
             sort_urls: sort results alphabetically by URL
             keep_source: per default, if selectors are defined, the HTML source will be discarded
        """
        await event_loop.create_task(self._crawl(event_loop))
        if len(self._results):
            self._process(sort_urls=sort_urls, keep_source=keep_source)

    @property
    def results(self) -> [dict]:
        return [page.__dict__ for page in self._results.values()]

    @property
    def errors(self) -> dict:
        return self._errors

    @property
    def cache(self):
        if self._cache is not None:
            return list(self._cache.values())
        raise ValueError('Cache is disabled')

    def clear_cache(self):
        if self._cache is not None:
            size = len(self._cache)
            self._cache.clear()
            LOG.info(f'Removed {size} items from cache')
        raise ValueError('Cache is disabled')

    def dump_cache(self, path: str = None):
        path = path or f'./dump-{self._domain}.json'
        with open(path, 'w') as file:
            file.write(json.dumps([page.__dict__ for page in self._cache.values()]))


if __name__ == '__main__':
    c = Microwler('https://quotes.toscrape.com/')
    c.run(verbose=True)



import asyncio
import logging
import os
from datetime import datetime

from quart import Quart, Response, send_from_directory
from quart_cors import cors

from microwler.utils import load_project, PROJECT_FOLDER

LOG = logging.getLogger(__name__)

app = Quart('Microwler')
app = cors(app, allow_origin='*')

STATIC = os.path.join(os.path.dirname(__file__), 'frontend/dist')
PROJECTS = dict()
STATUS = {
    'version': '0.1.8',
    'up_since': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
}


async def load_projects():
    global PROJECTS
    copy = dict()
    for path in os.listdir(PROJECT_FOLDER):
        if path.endswith('.py'):
            name = path.split('.')[0]
            project = load_project(name, PROJECT_FOLDER)
            if name in PROJECTS:
                PROJECTS[name]['start_url'] = project.crawler.start_url
                copy[name] = PROJECTS[name]
            else:
                copy[name] = {'name': name, 'start_url': project.crawler.start_url, 'last_run': dict()}
    PROJECTS = copy
    LOG.info(f'Imported {len(PROJECTS)} projects from filesystem')


@app.before_serving
async def init():
    await load_projects()


@app.route('/status')
async def status():
    """
    Return the service status

    - Route: `/status`
    - Method: `GET`
    - Response example:
    ```json
    {
        app: {
            up_since: "2021-02-05 17:42:13",
            version: "0.1.7"
        },
        projects: [
            "quotes"
        ]
    }
    ```
    """
    files = [file for file in os.listdir(PROJECT_FOLDER) if file.endswith('.py')]
    if len(files) != len(PROJECTS):
        await load_projects()

    return {
        'app': STATUS,
        'projects': list(PROJECTS.keys()),
    }


@app.route('/status/<project_name>')
async def project(project_name):
    """
    Return the project status

    - Route: `/status/<str:project_name>`
    - Method: `GET`
    - Response example:
    ```json
    {
        name: "quotes",
        start_url: "https://quotes.toscrape.com/"
        last_run: {
            state: "finished successfully",
            timestamp: "2021-02-05 17:47"
        },
    }
    ```
    """

    return PROJECTS[project_name]


@app.route('/crawl/<project_name>')
async def crawl(project_name: str):
    """
    Run the project's crawler and return the results

    - Route: `/crawl/<str:project_name>`
    - Method: `GET`
    - Response example:
    ```
    {
        data: [
            {
                url: "https://quotes.toscrape.com/"
                status_code: 200,
                depth: 0,
                discovered: "2021-02-05",
                links: [
                    "https://quotes.toscrape.com/tag/inspirational/",
                    "https://quotes.toscrape.com/author/Jane-Austen",
                    "https://quotes.toscrape.com/tag/obvious/page/1/",
                    "https://quotes.toscrape.com/tag/friends/",
                    "https://quotes.toscrape.com/tag/misattributed-eleanor-roosevelt/page/1/",
                    ...
                ],
                data: {
                    title: "Quotes to Scrape"
                    headings: {
                        h1: ["Quotes to Scrape"],
                        h2: ["Top Ten tags"],
                        h3: [""]
                    },
                },
            },
            ...
        ]
    }
    ```
    """
    PROJECTS[project_name]['last_run']['timestamp'] = datetime.now().strftime('%Y-%m-%d %H:%M')
    try:
        project = load_project(project_name, project_folder=PROJECT_FOLDER)
        loop = asyncio.get_event_loop()
        project.crawler.set_cache(force=True)
        await project.crawler.run_async(event_loop=loop)
        PROJECTS[project_name]['last_run']['state'] = 'finished successfully'
        return {'data': project.crawler.results}
    except Exception as e:
        LOG.error(e)
        PROJECTS[project_name]['last_run']['state'] = f'failed because: {e}'
        return Response(str(e), status=500)


@app.route('/data/<project_name>')
async def data(project_name: str):
    """
    Return the project's cached data

    - Route: `/data/<str:project_name>`
    - Method: `GET`
    - Response is in the same format as [above][microwler.web.backend.crawl]
    """
    project = load_project(project_name, project_folder=PROJECT_FOLDER)
    project.crawler.set_cache(force=True)
    cache = project.crawler.cache
    response = {'data': cache}
    return response


@app.route('/<folder>/<file>', methods=['GET'])
async def serve_folder(folder, file):
    return await send_from_directory(os.path.join(STATIC, folder), file)


@app.route('/<file>', methods=['GET'])
async def serve_file(file):
    return await send_from_directory(STATIC, file)


@app.route('/', methods=['GET'])
async def serve_index():
    return await send_from_directory(STATIC, 'index.html')


def start_app(port: int = 5000):
    """
    Starts the production-ready ASGI application using Hypercorn
    Arguments:
         port: the port to run on
    """
    from hypercorn.asyncio import serve
    from hypercorn.config import Config

    LOG.info('Starting webservice...')
    config = Config()
    config.bind = [f'localhost:{port}']
    config.loglevel = 'WARNING'
    try:
        LOG.info(f'Running on http://localhost:{port} (CTRL + C to quit)')
        asyncio.run(serve(app, config))
    except Exception as e:
        LOG.error(e)


if __name__ == '__main__':
    start_app()



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='microwler',
    version='0.1.8',
    description='A micro-framework for asynchronous deep crawls and web scraping written in Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://innovinati.github.io/microwler/',
    author='Maximilian Wolf',
    author_email='maximilian.wolf@innovinati.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Framework :: AsyncIO',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    packages=['microwler', 'microwler.cli', 'microwler.web'],
    include_package_data=True,
    python_requires='>=3.7, <4',
    install_requires=[
        'aiohttp', 'lxml', 'diskcache', 'prettytable', 'quart',
        'random-user-agent', 'html-text', 'completely', 'click',
        'parsel', 'chardet', 'quart-cors', 'aiodns'
    ],
    entry_points='''
    [console_scripts]
    microwler=microwler.cli.cmd:show_help
    new=microwler.cli.cmd:add_project
    crawler=microwler.cli.cmd:crawler
    serve=microwler.cli.cmd:start_server
    ''',
    project_urls={
        'Issues': 'https://github.com/INNOVINATI/microwler/issues',
        'Source': 'https://github.com/INNOVINATI/microwler/',
    },
)
